from __future__ import annotations

try:
    import pytest
except ImportError:
    pytest = None

import ast
import inspect
import textwrap
from typing import TypeAlias, Any
from enum import Enum, IntFlag, auto


class AppConfigException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class NameSpace:
    """Marker class for configuration namespaces.
    Used to group configuration variables into nested namespaces within an AppConfig subclass.
    See AppConfig documentation for more information.
    """

    pass


class PathType(Enum):
    Dir = auto()
    File = auto()


def get_attribute_docstrings(cls: type) -> dict[str, str]:
    """
    Extracts docstrings for class attributes by parsing the source code.
    This allows using docstrings to populate ConfigVarDef.help.

    currently PEP257 is not yet implemented in python standard library so we need to parse the source code ourselves.
    This can slow down the class creation but this is done at class creation time only.
    """
    docstrings = {}
    try:
        source = inspect.getsource(cls)
        # Deduct to handle nested classes indentation
        source = textwrap.dedent(source)
    except (OSError, TypeError):
        # Source code not available (e.g. dynamic class, REPL, compiled files)
        return {}

    try:
        tree = ast.parse(source)
    except SyntaxError:
        return {}

    # We look for the ClassDef in the parsed source.
    # Since inspect.getsource(cls) returns the class definition itself,
    # the first node in body should be the ClassDef.
    for node in tree.body:
        if isinstance(node, ast.ClassDef):
            for i, item in enumerate(node.body):
                if isinstance(item, (ast.Assign, ast.AnnAssign)):
                    # Check if the next node is an expression containing a string (docstring)
                    if i + 1 < len(node.body):
                        next_node = node.body[i + 1]
                        if (
                            isinstance(next_node, ast.Expr)
                            and isinstance(next_node.value, ast.Constant)
                            and isinstance(next_node.value.value, str)
                        ):
                            docstring = next_node.value.value.strip()

                            targets = []
                            if isinstance(item, ast.Assign):
                                targets = item.targets
                            elif isinstance(item, ast.AnnAssign):
                                targets = [item.target]

                            for target in targets:
                                if isinstance(target, ast.Name):
                                    docstrings[target.id] = docstring
            # We found the class def, no need to continue in top level
            break

    return docstrings


def is_sequence(obj: Any) -> bool:
    """test if obj is a sequence (list, tuple, etc...) but not a string"""
    try:
        len(obj)
        obj[0:0]
        return not isinstance(obj, str)
    except KeyError:
        return False
    except TypeError:
        return False  # TypeError: object is not iterable


class _UndefinedSentinel:
    pass


Undefined: TypeAlias = _UndefinedSentinel
_undef: Undefined = _UndefinedSentinel()


if pytest:
    pass



from __future__ import annotations

from collections.abc import Callable, Sequence
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Optional,
    TypeVar,
    Union,
    cast,
    dataclass_transform,
    get_args,
    get_origin,
    get_type_hints,
)
import inspect
from .utils import Undefined, _undef, PathType, is_sequence
from .flags import Flags

if TYPE_CHECKING:
    from .app_config import AppConfig

_T = TypeVar("_T")


@dataclass_transform()
class ConfigVarDef(Generic[_T]):
    """
    Definition of an application configuration variable.

    Attribute of this class are called configuration directives that are used to specify the configuration of the application with the AppConfig class.

    Configuration are used indirectly by defining attributes of an AppConfig subclass with the configvar() function that creates ConfigVarDef objects:
        class MyAppConfig(AppConfig):
            my_var: int = configvar(default=1)

    The method resolve_vars() of the AppConfig class uses these configuration variable definitions to look for the values of the variables as explained in its documentation.
    Each configuration variable definition contains several directives to define how to look for the values of the variables in a dictionnarary that can be the CLI parameters,
    or a configuration file, or an environment variables or a computation defined by a python Callable.

    Two directive are not specified in the constructor of ConfigVarDef because they are inferred from the attribute name and type hint:
        - name : the name of the variable, inferred from the attribute name.
        - type_info : the type to which to cast the variable value, inferred from the attribute type hint. Must be a python builtin type or a class or a list of
            builtin types.
            Here how to get the list of python builtin types:
                print([t.__name__ for t in __builtins__.__dict__.values() if isinstance(t, type)])
            Lists are handle with a specific processing detail below in the paragraph 'Lists'.
            Complex casting can be handle with a transfom Eval Form or with a class that implements the complex type.
            Optional value are supported like Optional[int] or Union[str, None] to indicate that the variable can be None.
            When the type hint is missing the variable is considered to be of type str.

    The directives that can be specified in configvar() are:

    - help [optional]: a help string that describes the variable.
    - default [optional]: the default value, can be a value or an Eval Form (see below explanation on Eval forms) whose
        returned value will be the value of the variable if no other value has been found.
        See paragraph about None below for more information.
    - env_name [optional ]: the name of the environment variable that can contain the value. By default search for name
    - file_key [optional ]: the key to look for in the configuration file, key can be in the form a1.a2.a3 (attributes separated by dots) to indicate
        nested namespaces (see AppConfig documentation). By default search for name.
    - value_key [optional ]: the key to look for in the "Values" dict. by default search for name. None value are ignored.
        If the constructor parameter click_key_conversion flag (see below for flags) is true then the default value is the name in lowercase and with '-' replaced by '_'
    - transform : an Eval Form (see below explanation on Eval forms) that transform the value found for the variable and produce another value that will be the value
        of the variable before possible casting (like explain in type_info).
    - scopes [optional]: a list of names indicating in which scope the variable is valid. If one of the scope names passed at variable resolution to
        AppConfig.resolve_vars() matches one of the scope names of the variable or if the variable has no scope directive then the variable is
        considered for resolution.
    - split_to_list [optional ]: Indicates that the value found must be split in order to produce a list of string. The value

        of the directive is the list separator or True if the separator is a comma. Value in the string list will be casted to the type specified in type_info.
        See the paragraph below about Lists that explain how list handled.
    - can_be_relative_to [optional]: a directory path or the name of another key that resolves in a directory path.
        If type_info is a Path and the retrieved value is not an empty string nor None nor an absolute path then
        transform the value to an absolute path with the given directory as parent directory.
        See details below on how Path are processed.
    - make_dirs [optional ]: a PathType that indicates the type of Path for which directories are to be created if they do not exist yet.
            if PathType is:
                - Dir : create the hierarchy of directory necessary to contains this directory as well as the directory itself
                - File: create the parent hierarchy of directory necessary to contains this file
        See details below on how Path are processed.
    - auto_prolog [optional]: mark the variable for automatic prolog processing - which means that the variable will be resolved at an early stage, before the configuration file
        is resolved and before all other variables. If it has also the allow_override flag True then it can be processed again with all other variables in the regular resolution.


    - Flag directives:
        These directives can also be defined globally in the AppConfig object or when calling resolve_vars().Flags defined in a ConfigVarDef override
        the other values if they are defined.
        - no_dir_processing [optional ]: don't run specific Path processing for this variable.
        - no_env_search [optional ]: boolean, False by default. If true the value of the variable is not searched in the Environment.
        - no_value_search[optional ]: boolean, False by default. If true the value of the variable is not searched in the values dict.
        - no_conffile_search [optional ]: boolean, False by default. If true the value of the variable is not searched in the configuration file.
        - click_key_conversion [optional ]: boolean, False by default. If true the value of the variable is converted using Click key conversion.
        - allow_override [optional ]: boolean, False by default. If true the variable value can be overridden by another context.
        - mandatory [optional ]: boolean, True by default. If true an exception is thrown if the variable value is not found. Otherwise the variable is not set which means that accessing it
          would raise an Exception.
        - no_search [optional] : boolean, False by default. It true the value of the variable is not searched in the Environment, the values or the configuration
            file. It's equivaleut to set the directives no_env_search, no_value_search, no_conffile_search to True. In this case the value of the variable should be defined by
            the default directive.

    For backward compatibilty the special value __NO_SEARCH__ for env_name, config_key, file_key and value_key are still supported but it's recommended to use
    the no_env_search, no_value_search, no_conffile_search or no_search directives instead.

    By default throw an exception if no value are found except if mandatory is False

    Eval Forms:
    -----------

        an eval form is a callable that is evaluated to compute a value for a variable. Eval forms can be used in the default directive or the
        transform directive.

        The callable receives three parameters: the variable name, this configuration object and the current value found for the variable if one has been found
        in the case of a transform directive.

        The expression can use value already defined in this application configuration object by using the first parameter of the callable which is the AppConfig object itself,
        knowing that variables are evaluated in the order they are defined.

        In the case of a default directive the third parameter of the callable will be None.
        In the case of a transform directive the third parameter of the callable may be None if no value have been found and
        the 'mandatory' global flag is set to False, otherwise it will be the value found for the variable before transformation.

        The returned value from the evaluation of the callable will be used to set the variable value before casting to type_info.
        Note that the eventual type casting will be done after the Eval call.
        It's up to the caller to deal with the returned value.

        Note that the libraries used in Eval Forms must be imported in the module that defines the form.

        !!Warning!!: default and transform directives should be set or return a value that must be 'castable' to the type defined by type_info.
        For example if a variable is declared as a class that accept a string in its constructor then the default directive should be a string or an Eval
        Form that returns a string not the class instance.

    Path and Path list type_info variables:
    --------------------------------------

        If a variable has a type_info that is a Path or a list of Path there a dedicated processing that applies on the
        value found for them.
        This processing takes place at the end of the processing of normal variable and occurs when:
            - the type_info is a Path or a list of Path
            - a value for the variable was found
            - the directive no_dir_processing is False

        The specific processing is the following on the Path or on each of the Path in the list of Path:
            - can_be_relative_to: if this directive has a value that can be a path-like directory or the name of another key already defined
            that resolves in a path-like directory and the value found is not an absolute Path:
                - then make a Path with the can_be_relative_to directory as parent and the value found as its child.
            - make_dirs: if this directive is True , then create the necessary directories of the path if they don't exist yet.
            - Call the 'expanduser()' and 'resolve()' method on the resulting Path.

    List:
    -----

        All values found are strings that should be casted to the type specified in type_info inferred from the type hint of the variable.
        Only list of single type are supported like list[int], list[str], list[MyClass], etc.
        The value is considered as a list when split_to_list is not False and a value was found for the variable.
        In this case the temporaray string value found for the variable is split with the string 'split()' method to generate
        a list of string.
        Then all items of the list of strings are casted to the specified type_info.

    scopes:
    --------
        Every configuration variable can specify a list of scope(s) in which it is valid.
        When resolving variables it is also possible to specify for which scope(s) the resolution must be done.

        When one or several scopes are specify for the resolution then only the variable valid for these scopes will be considered for resolution
        as well as the VarDef without Scope directive. Variable defined without Scope directive are common to all scope, the one with a Scope directive
        are specific to their scopes.

        This allows to define variables that are specific to some scope and to resolve only these variables when needed.
        When the flage allow_override is False and several resolutions are done, then variables are evaluated only once even if they are valid in several
        of the requested scope(s), only the first valid scope is used.

    None value:
    ----------
        All the VarDef valid for the involved scopes will be resolved and the one for which no value have been found will be set to None if the
        flag 'mandatory' is False.

        If the flag 'mandatory' is true and one varDef for the scope has no value then an exception is raised.

        Whatever is the value of the flag 'mandatory' it is possible to set a value to None either with the 'default' directive or with
        an Eval Form or a None value in one of the search location.

    AutoProlog variables:
    --------------------
        Variables defined with the auto_prolog directive set to True are considered as AutoProlog variables. These variables are resolved at an early stage
        before all other variables and before the configuration file is processed.
        This allows to define variables that are necessary to locate the configuration file when its path depends on other variables. In this case these variables
        must be defined as AutoProlog variables.
        The default variables defined in the AutoProlog class are all AutoProlog variables and therefore are made available to be used in variable substitution for all other
        variables and for locating the configuration file.

    Note on implementation:
    -----------------------
        ConfigVarDef is a descriptor class that defines the __get__ and __set__ methods.
        This allows to define configuration variables as class attributes on AppConfig subclasses.
        The __set_name__ method is used to set the name and type_info directives based on the attribute name and type hint.
    """

    _auto_prolog: Optional[bool] = None
    _can_be_relative_to: Optional[Path | str] = None
    _scopes: Optional[Sequence[str]] = None
    _default: Any | Callable[[str, Any, Any], Any] | Undefined = _undef
    _env_name: Optional[Sequence[str]] = None
    _file_key: Optional[Sequence[str]] = None
    _help: str = ""
    _make_dirs: Optional[PathType] = None
    _mandatory: Optional[bool] = True
    _no_dir_processing: Optional[bool] = None
    _name: Optional[str] = None
    _split_to_list: Optional[bool | str] = None
    _transform: Optional[Union[Callable[[str, Any, Any], Any], str]] = None
    _type_info: Any = None
    _value_key: Optional[Sequence[str]] = None

    def __init__(
        self,
        *,
        allow_override: Optional[bool] = None,
        auto_prolog: Optional[bool] = None,
        can_be_relative_to: Optional[Path | str] = None,
        click_key_conversion: Optional[bool] = None,
        scopes: Optional[Sequence[str]] = None,
        default: Any | Callable[[str, Any, Any], Any] | Undefined = _undef,
        env_name: Optional[str | Sequence[str]] = None,
        file_key: Optional[str | Sequence[str]] = None,
        help: str = "",
        make_dirs: Optional[PathType] = None,
        mandatory: Optional[bool] = None,
        no_conffile_search: Optional[bool] = None,
        no_dir_processing: Optional[bool] = None,
        no_env_search: Optional[bool] = None,
        no_search: Optional[bool] = None,
        no_value_search: Optional[bool] = None,
        split_to_list: Optional[bool | str] = None,
        transform: Optional[Union[Callable[[str, Any, Any], Any], str]] = None,
        type_info: Any = None,
        value_key: Optional[str | Sequence[str]] = None,
    ) -> None:
        """Initialize the configuration variable definition.
        Should not be used directly but be used through the function configvar() to define a configuration variable as a class attribute
        on an AppConfig subclass like this:
            class MyCfg(AppConfig):
                my_var: int = configvar(default=1)

        """
        self.__doc__ = help
        self._auto_prolog = auto_prolog
        self._can_be_relative_to = can_be_relative_to
        self._scopes = scopes
        self._default = default
        self._env_name = (
            None
            if env_name is None
            else (
                env_name
                if is_sequence(env_name)
                else [env_name] if isinstance(env_name, str) else None
            )
        )
        self._file_key = (
            None
            if file_key is None
            else (
                file_key
                if is_sequence(file_key)
                else [file_key] if isinstance(file_key, str) else None
            )
        )
        self._help = help
        self._make_dirs = make_dirs
        self._mandatory = mandatory
        self._no_dir_processing = no_dir_processing
        self._no_search = no_search
        self._split_to_list = split_to_list
        self._transform = transform
        self._type_info = type_info
        self._value_key = (
            None
            if value_key is None
            else (
                value_key
                if is_sequence(value_key)
                else [value_key] if isinstance(value_key, str) else None
            )
        )
        self.flags = Flags(
            no_env_search=no_env_search,
            no_value_search=no_value_search,
            no_conffile_search=no_conffile_search,
            click_key_conversion=click_key_conversion,
            allow_override=allow_override,
        )

    def __set_name__(self, owner: type, name: str) -> None:
        """This method is a special python method called when a descriptor is assigned to a class attribute.
        ConfigVarDef is a descriptor class because it defines the __get__ and __set__ methods.

        The goal of this method is to:
            - set the name directive of this ConfigVarDef based on the attribute name.
            - infer the type of the attribute and set the type_info directive with it.

        Args:
            owner (type): the owning class of the attribute that is being assigned the descriptor
            name (str): the attribute name
        """

        # set name directive
        self._name = name

        # try to find a type:
        #     if we have a type hints and a type_info directive we take the type_info.
        #     This allow to define a type for the type checker that allow override with another type if the user wants to.
        #     (it is usefull for exec_stage)
        #     but still have a correct run time type , this is correct as long as the user use duck typing types
        #     that are compatible.
        try:
            hints = get_type_hints(owner, include_extras=True)
        except Exception:
            hints = getattr(owner, "__annotations__", {})

        # test if there is a type hint for this attribute
        annotated_type: dict[str, Any] | Any | dict[Any, Any] = None
        if name in hints:
            annotated_type = hints[name]
            if isinstance(annotated_type, str):
                # Why can it be a string? In Python, type hints are strings if:
                #   from __future__ import annotations is used (postponed evaluation of annotations).
                #   The user explicitly used quotes for a forward reference (e.g., var: "MyClass" where MyClass is defined later).
                for frame_info in inspect.stack():
                    # try to resolve the type with an eval to which we provide the globals and locals for its resolution.
                    # But it can happen that the type is defined in a local scope, so we need to go through all frames
                    # using the stack.
                    try:
                        resolved_type = eval(
                            annotated_type,
                            frame_info.frame.f_globals,
                            frame_info.frame.f_locals,
                        )
                        if resolved_type is not None:
                            annotated_type = resolved_type
                            break
                    except Exception:
                        continue
                else:
                    raise TypeError(
                        f"Could not resolve type hint '{annotated_type}' for configuration variable '{name}'. "
                        "Please check that the type is imported and available."
                    )

        type_info_directive: type | None = None
        if self._type_info is not None:
            type_info_directive = self._type_info

        if annotated_type is not None and type_info_directive is not None:
            # both type hint and type_info directive are defined, use type_info directive
            self._type_info = type_info_directive
        elif annotated_type is not None:
            # only type hint is defined, use it
            self._type_info = annotated_type
        elif type_info_directive is not None:
            # only type_info directive is defined, use it
            self._type_info = type_info_directive
        else:
            # no type hint nor type_info directive, default to str
            self._type_info = str

        # We still need to handle split_to_list if not set and the type selected is a list
        if self._split_to_list is None:
            try:
                origin = get_origin(self._type_info)
                args = get_args(self._type_info)

                # Check for basic list
                if origin is list or self._type_info is list:
                    self._split_to_list = True
                # Check for Optional[list] / Union[list, ...]
                elif origin is Union:
                    for arg in args:
                        arg_origin = get_origin(arg)
                        if arg_origin is list or arg is list:
                            self._split_to_list = True
                            break
            except Exception:
                pass

    def __get__(
        self, instance: Optional[AppConfig], owner: Optional[type] = None
    ) -> ConfigVarDef[_T] | _T | None:
        """
        Descriptor method that gets the value of the configuration variable from the AppConfig instance.

        The value is retrieved from the instance's __dict__ using the variable's name because we want to
        handle the case where we have dotted names for nested attributes.
        """
        if instance is None:
            return self
        name = self.name
        if name in instance.__dict__:
            return instance[name]
        return None

    def __set__(self, instance: AppConfig, value: Any) -> None:
        """
        Descriptor method that gets the value of the configuration variable from the AppConfig instance.
        The value is set in the instance's __dict__ using the variable's name because we want to
        handle the case where we have dotted names for nested attributes.
        """
        instance._set_value(self.name, value)

    @property
    def auto_prolog(self) -> Optional[bool]:
        return self._auto_prolog

    @property
    def can_be_relative_to(self) -> Optional[Path | str]:
        return self._can_be_relative_to

    @property
    def scopes(self) -> Sequence[str] | None:
        return self._scopes

    @property
    def default(self) -> Any | Callable[[str, Any, Any], Any] | Undefined:
        return self._default

    @property
    def env_name(self) -> Optional[Sequence[str]]:
        if self.flags.no_search or self.flags.no_env_search:
            return None
        else:
            return self._env_name or self.name

    @property
    def file_key(self) -> Optional[str | Sequence[str]]:
        if self.flags.no_search or self.flags.no_conffile_search:
            return None
        else:
            return self._file_key or self.name

    @property
    def help(self) -> str:
        return self._help

    @property
    def make_dirs(self) -> Optional[PathType]:
        return self._make_dirs

    @property
    def mandatory(self) -> bool:
        return True if self._mandatory is None else self._mandatory

    @property
    def no_dir_processing(self) -> bool:
        return False if self._no_dir_processing is None else self._no_dir_processing

    @property
    def name(self) -> str:
        if self._name is None:
            raise RuntimeError(
                "ConfigVarDef has no name. Declare it on an AppConfig subclass "
            )
        return self._name

    @property
    def split_to_list(self) -> bool | str:
        return False if self._split_to_list is None else self._split_to_list

    @property
    def transform(self) -> Optional[Union[Callable[[str, Any, Any], Any], str]]:
        return self._transform

    @property
    def value_key(self) -> Optional[str | Sequence[str]]:
        """Note: we can't compute the final value here because it depends on the click_key_conversion
        flag that can be set globally or in resolve_vars."""
        if self.flags.no_search or self.flags.no_value_search:
            return None
        else:
            return self._value_key or self.name


def configvar(
    *,
    allow_override: Optional[bool] = None,
    auto_prolog: Optional[bool] = None,
    can_be_relative_to: Optional[Path | str] = None,
    click_key_conversion: Optional[bool] = None,
    scopes: Optional[Sequence[str]] = None,
    default: Any | Callable[[str, Any, Any], Any] | Undefined = _undef,
    env_name: Optional[str | Sequence[str]] = None,
    file_key: Optional[str | Sequence[str]] = None,
    help: str = "",
    make_dirs: Optional[PathType] = None,
    mandatory: Optional[bool] = None,
    no_conffile_search: Optional[bool] = None,
    no_dir_processing: Optional[bool] = None,
    no_env_search: Optional[bool] = None,
    no_search: Optional[bool] = None,
    no_value_search: Optional[bool] = None,
    split_to_list: Optional[bool | str] = None,
    transform: Optional[Union[Callable[[str, Any, Any], Any], str]] = None,
    type_info: Any = None,
    value_key: Optional[str | Sequence[str]] = None,
) -> Any:
    """Create a declarative configuration variable definition.

    Use it as a class attribute on an AppConfig subclass:

        class MyCfg(AppConfig):
            my_var: int = configvar(default=1)

    The variable name defaults to the attribute name and its type is inferred from the type hint of the attribute declaration.

    This function allows to assign a ConfigVarDef objects to an attribute definitions of any type by telling the type checkers
    that the type of the assignation is Any instead of ConfigVarDef.
    Indeed ConfigVarDef is a generic class and my_var: int = ConfigVarDef(...) or my_var: ConfigVarDef[int] = ConfigVarDef(...)
    would not be accepted by type checkers. configvar() declares returning Any which deactivates type checking.

    """

    return ConfigVarDef(
        allow_override=allow_override,
        auto_prolog=auto_prolog,
        can_be_relative_to=can_be_relative_to,
        click_key_conversion=click_key_conversion,
        scopes=scopes,
        default=default,
        env_name=env_name,
        file_key=file_key,
        help=help,
        make_dirs=make_dirs,
        mandatory=mandatory,
        no_conffile_search=no_conffile_search,
        no_dir_processing=no_dir_processing,
        no_env_search=no_env_search,
        no_search=no_search,
        no_value_search=no_value_search,
        split_to_list=split_to_list,
        transform=transform,
        type_info=type_info,
        value_key=value_key,
    )



from __future__ import annotations

import jinja2
from typing import Any, Callable, Mapping, Optional, Sequence, Union, cast
from .utils import is_sequence, _undef

RenderContext = Mapping[str, Any]


def render_template(value: Union[str, Any], context: RenderContext) -> Any:
    """
    Render a string as a jinja2 template using the provided context.
    If value is not a string, return it as is.
    """
    if isinstance(value, str):
        return jinja2.Template(value).render(context)
    return value


def resolve_candidate_list(
    *,
    candidates: Optional[Union[str, Sequence[str]]],
    context: RenderContext,
    check_exists: Optional[Callable[[str], Any]] = None,
) -> Optional[Any]:
    """
    Resolve a list of candidate strings (which can be Jinja2 templates) to find the first valid one.

    Args:
        candidates: A single string or a list of strings (potentially templates).
        context: The context used for rendering templates.
        check_exists: An optional predicate to validate the rendered string.
                      returns _undef if not valid, or the value to return if valid.
    Returns:
        The first successfully resolved and validated candidate string or the first not None value return by check_exists, or None if no match is found.
    """
    if candidates is None:
        return None

    candidates_list: list[str] = []
    if isinstance(candidates, str):
        candidates_list = [candidates]
    elif is_sequence(candidates):
        candidates_list = cast(list[str], candidates)

    for raw_candidate in candidates_list:
        rendered = render_template(raw_candidate, context)
        if rendered is None:
            continue

        if isinstance(rendered, str):
            if check_exists is not None:
                res = check_exists(rendered)
                if res != _undef:
                    return res
            else:
                return rendered

    return _undef



from __future__ import annotations

import copy

import json
import logging
import os

from collections import OrderedDict
from collections.abc import Iterable, Mapping, Sequence
from pathlib import Path
from typing import Any, Callable, Optional, Union, cast

from .auto_prolog import AutoProlog
import toml
import yaml
from pydantic import TypeAdapter, ValidationError

from .attr_dict import AttrDict
from .utils import is_sequence, Undefined
from .config_var import ConfigVarDef, Flags
from .solver import Solver
from .var_substitution import resolve_candidate_list
from .utils import _undef
from .app_config_meta import AppConfigMeta, ConfigVarCollector


class AppConfig(AttrDict, AutoProlog, metaclass=AppConfigMeta):
    """ "
    A class to manage application configuration. It can be passed from callable to callable.
    It can search for variable values in command line arguments, environment variables, a yaml or json configuration file
    or a default value or a computed value.

    Configuration variable are defined with a list of ConfigVarDef object which defines how to look for the variable values and/or compute them.

    The variable values are searched in these locations in the following order:
    - a dict of values usually passed from the command line arguments
    - environment variables
    - a configuration file in yaml or json format
    - a default value or a computed value
    The first match found is used as the value for the variable.

    When found the variable can be casted to a specific type and/or transformed with a callable.

    The default value can also be computed with a callable that can use already defined variables in this AppConfig object.
    One can combine a default value and a transformation callable to compute complex default values that can combine several already defined variables,
    knowing that variables are resolved in the order they are declared in the AppConfig subclass.

    See method 'resolve_vars()' and class 'ConfigVarDef' for more information.

    Variable can also be set directly like in an object attribute or in a Dict key:

    ctx = AppConfig()
    ctx["var"] = 2
    print(ct.var) # print 2

    Accessing the variables can also be done like an object attribute or a Dict key:
    ctx = AppConfig()
    ctx.var = 2
    print(ctx["var"]) # print 2

    This can be useful to mix known variable name and variable dynamically defined in a file or passed from the command line.
    def foo(ctx: AppConfig,attname:Any) -> Any:
        return ctx[attname]

    Variable can contain variable recursively, in this case they can be access with a dotted notation or with a dict like way:
    ctx = AppConfig()
    ctx["var1.var2"] = 2
    ctx["var1.var3"] = 3
    print(f"var1 = {ctx.var1}") # print a dict {"var2":2,"var3":3}

    print(ctx.var1.var2) # print 2

    Declarative Configuration:
    --------------------------
    Configuration is best defined declaratively by subclassing AppConfig:

        class MyConfig(AppConfig):
            host: str = configvar(default="localhost")
            port: int = configvar(default=8080)

    Nested Namespaces:
    ------------------
    Variables can be grouped into namespaces using nested classes inheriting from NameSpace.
    This creates a structured configuration:

        class MyConfig(AppConfig):
            # Root level variable
            debug: bool = configvar(default=False)

            # Nested namespace 'db'
            class db(NameSpace):
                host: str = configvar(default="db.local")
                port: int = configvar(default=5432)

            # Nested namespace 'api'
            class api(NameSpace):
                key: str = configvar(env_name="API_KEY")

    inst = MyConfig()
    inst.resolve_vars()
    print(inst.db.host)  # Access nested variables

    Decentralized Configuration (Mixins):
    -------------------------------------
    Configuration can be split across multiple classes (Mixins) and combined using inheritance.
    This allows different modules to define their own configuration requirements.

        class DatabaseConfig(AppConfig):
            class db(NameSpace):
                host: str = configvar(default="localhost")

        class ApiConfig(AppConfig):
            class api(NameSpace):
                timeout: int = configvar(default=30)

        # Combine into the final application config
        class App(DatabaseConfig, ApiConfig):
            pass

    If multiple mixins define the same namespace (e.g. `class db(NameSpace)`), their variables are merged.
    """

    _logger = logging.getLogger(__name__)
    _global_flags: Flags
    _config_var_defs: OrderedDict[str, ConfigVarDef]
    """ The list of ConfigVarDef defined in this AppConfig class and its parent classes as an ordered dictionary. 
        This attribute is created at class creation time by the __init_subclass__ method.
    """
    _autoprolog_var_names: list[str]
    """ The list of the names of ConfigVarDef defined in AutoProlog.
        This attribute is created at class creation time by the __init_subclass__ method.
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Call when a subclass inherits from AppConfig.

        Create the _config_var_defs attribute that contains the list of ConfigVarDef
        defined in this subclass as well as the ones defined in parent classes.

        This method also handle namespaces defined with nested classes inheriting from NameSpace.
        1. It first collects the _config_var_defs from parent classes to handle mixins, when a variable is found in a parent class with the same name as a variable in a child class,
            the child class variable takes precedence, thus we remove the parent class variable from the collected list before adding the child class variable.
        2. Then it inspects the class __dict__ to find ConfigVarDef defined in this class (and nested classes) and add them to the collected list.
            Collecting ConfigVarDef in this class is done by the collect_vars() class method which is called recursively on nested NameSpace classes
            and which is also search for docstrings to attach them as help to the ConfigVarDef if help is not already defined.
        3. Finally it sets the _config_var_defs attribute on the class.

        """
        # First call super to handle any parent class logic
        super().__init_subclass__(**kwargs)

        with ConfigVarCollector() as collector:
            # try to look if there is already a _config_var_defs attribute in one of the parent classes.

            for base in reversed(cls.__mro__[1:-1]):
                # __mro__ contain:
                #    1. The class of the object (class of cls here, which is this class)
                #    2. The hierarchy of the parent classes in the MRO order which goes from left to right at each level and from the level
                #       closest to the object to one deeper in the tree.
                #    3. The Object class which is the latest class in the inheritance tree.
                # Here we skip the first element which is the class of the object itself (cls) and the last which is Object class and we traverse
                # it in reverse order.

                collector.collect_in_config_var_defs(base)

            # Now inspect the class __dict__ to find ConfigVarDef defined in this class (and nested classes) and add them to the collected list.
            # note that collect_config_vars is defined in the metaclass.
            collector.collect_on_class(cls)
            cls._config_var_defs = collector.get_collected_vars()
            cls._autoprolog_var_names = collector.get_auto_prolog_var_names()

    def __init__(
        self,
        *,
        no_env_search: Optional[bool] = None,
        no_key_search: Optional[bool] = None,
        no_conffile_search: Optional[bool] = None,
        no_search: Optional[bool] = None,
        allow_override: Optional[bool] = None,
        click_key_conversion: Optional[bool] = None,
    ):
        """Init the config.

        Arguments are called 'global flags', they specify the behavior of the variable resolution globaly. They can be overiden
        when the resolve method is call. (this method can be call several time and thus on each call some flags can be changed) or
        per variable when defining the ConfigVarDef object.

        Global Flags:
            no_env_search (bool, optional): don't search values in the environment. Defaults to False.
            no_key_search (bool, optional): don't search values in the value dictionary. Defaults to False.
            no_conffile_search (bool, optional): don't search values in the configuration file. Defaults to False.
            no_search (bool, optional): don't search values in any location. Defaults to False.
            allow_override (bool, optional): allow overriding variable in the configuration when calling resolve_vars.
                it happens either when resolve_vars() is call several times or the variable was set before its call.
            click_key_conversion (bool, optional): use the variable name converted to lower case and with dashes converted to underscores as
            the key to search in values. Defaults to False.
        """
        self._global_flags = Flags(
            no_env_search=no_env_search,
            no_value_search=no_key_search,
            no_conffile_search=no_conffile_search,
            no_search=no_search,
            click_key_conversion=click_key_conversion,
            allow_override=allow_override,
        )

        self._logger.debug("init AppConfig")

    def clone(self) -> AppConfig:
        """cloning the AppConfig object"""
        clone = AppConfig()
        clone.__dict__ = self.__dict__.copy()
        return clone

    def _set_value(self, name: str, value: Any) -> None:
        """Set a value.

        Use an AttrDict implementation to handle dotted keys and not trigger recursion loops
        if called from a descriptor.
        """
        self[name] = value

    def _get_eval_runner(
        self,
    ) -> Callable[[Callable[[str, Any, Any], Any], str, Any], Any]:
        """
        Call a Callable found in a Eval Form.
        """
        this = self

        def runner(
            callable: Callable[[str, Any, Any], Any], var_name: str, var_val: Any
        ) -> Any:
            return callable(var_name, this, var_val)

        return runner

    def _resolve_conffile_path(
        self, candidates: Optional[str | Path | Iterable[str | Path]]
    ) -> Path | None:
        if candidates is None:
            return None

        # Normalize candidates to list of str for resolve_candidate_list
        candidates_str_list: list[str] = []
        if isinstance(candidates, (str, Path)):
            candidates_str_list = [str(candidates)]
        elif is_sequence(candidates):
            candidates_str_list = [str(c) for c in cast(Sequence[Any], candidates)]

        # Define existence check
        def file_exists(p: str) -> Any:
            path_obj = Path(p)
            if path_obj.is_file():
                return path_obj
            else:
                return _undef

        # Build context for Jinja with current values resolved
        context = self.to_dict()

        # Resolve using current self as context (for jinja variables)
        resolved_path = resolve_candidate_list(
            candidates=candidates_str_list, context=context, check_exists=file_exists
        )

        return resolved_path

    def resolve_vars(
        self,
        *,
        scopes: Optional[str | Iterable[str]] = None,
        values: Mapping[str, Any] = {},
        no_env_search: Optional[bool] = None,
        no_value_search: Optional[bool] = None,
        no_conffile_search: Optional[bool] = None,
        no_search: Optional[bool] = None,
        conffile_path: Optional[Path | str | Sequence[str | Path]] = None,
        allow_override: Optional[bool] = None,
        click_key_conversion: Optional[bool] = None,
    ) -> None:
        """
        Resolve the configuration variables defined in this AppConfig object.
        The variable values are searched in these locations in the following order:
        - a dict of values usually passed from the command line arguments
        - environment variables
        - a configuration file in yaml or json format
        - a default value or a computed value
        The first match found is used as the value for the variable.
        When found the variable can be casted to a specific type and/or transformed with a callable.
        The default value can also be computed with a callable that can use already defined variables in this AppConfig object.
        One can combine a default value and a transformation callable to compute complex default values that can combine several already defined variables,
        knowing that variables are resolved in the order they are declared in the AppConfig subclass.
        """

        self._logger.info(
            f"Starting variable resolution, conf file = {str(conffile_path)}  "
        )

        # Compute flags that can be override locally.
        resolve_flags = Flags(
            no_env_search=no_env_search,
            no_value_search=no_value_search,
            no_conffile_search=no_conffile_search,
            no_search=no_search,
            click_key_conversion=click_key_conversion,
            allow_override=allow_override,
        )
        local_flags = self._global_flags.merge(inplace=False, other=resolve_flags)
        assert local_flags is not None

        # building the Solver who is in charge of resolving configuration variables
        solver = Solver(
            eval_runner=self._get_eval_runner(),
            flags=local_flags,
            values_map=values,
            var_solved=self,
        )

        # Special case for auto_prolog variable: we need to try to solve them
        # before any other variable and before resolving the configuration file path because
        # it can depends on one of them, the one that can be ovveriden are resolved at each run
        for var_name in self._autoprolog_var_names:
            vardef = self._config_var_defs.get(var_name)
            if vardef is None:
                continue
            if self.has_value(var_name) and not (
                local_flags.allow_override or vardef.flags.allow_override
            ):
                continue
            # call solver to try to find a value for the configuration variable
            value_found = solver.resolve_confvar(var_to_solve=vardef, scopes=scopes)
            if value_found != _undef:
                self._set_value(var_name, value_found)

        # Read configuration file if requiered
        conf_file_vars = None
        if not local_flags.no_conffile_search and conffile_path is not None:
            # Resolve conffile_path if it's dynamic
            cur_conffile_path = self._resolve_conffile_path(conffile_path)

            if cur_conffile_path is not None and cur_conffile_path.is_file():
                with cur_conffile_path.open("r") as f:
                    if cur_conffile_path.suffix.lower() == ".json":
                        self._logger.debug(f"Parsing json file {cur_conffile_path}")
                        conf_file_vars = json.load(f)
                    elif cur_conffile_path.suffix.lower() == ".toml":
                        self._logger.debug(f"Parsing toml file {cur_conffile_path}")
                        conf_file_vars = toml.load(f)
                    else:
                        self._logger.debug(
                            f"Parsing file {cur_conffile_path} with suffix {cur_conffile_path.suffix.lower()}"
                        )
                        conf_file_vars = yaml.safe_load(f)
                    # updating the solver with the values found in the configuration file
                    solver.set_conffile_values_map(conf_file_vars)
            else:
                self._logger.info(f"configuration file {cur_conffile_path} not found.")

        # Read the var definitions one by one in their definition order
        for var_name, var_def in self._config_var_defs.items():

            # Check if variable is manually set on the instance or resolved in a previous pass
            has_value = self.has_value(var_name)

            # compute the current flags for this variable by merging local flags with variable flags
            cur_flags = local_flags.merge(inplace=False, other=var_def.flags)
            assert cur_flags is not None

            # if the variable is already defined and override is not allowed then
            # we skip it (First valid scope wins).
            if not cur_flags.allow_override and has_value:
                continue

            # call solver to try to find a value for the configuration variable
            value_found = solver.resolve_confvar(var_to_solve=var_def, scopes=scopes)
            if value_found != _undef:
                self._logger.info(
                    f"Variable '{var_name}' resolved to: {str(value_found)}"
                )
                # add the found value to the AppConfig object through the AttrDict implementation.
                self._set_value(var_name, value_found)
            else:
                self._logger.debug(f"Variable '{var_def.name}' could not be resolved.")

    def _serialize_var_defs(
        self, vardef_to_serialize: Iterable[ConfigVarDef], for_file: bool = False
    ) -> dict[str, Any]:
        """
        Serialize the given variable definitions to a dictionary.
        We know that values came from strings and that they are either alone or in a list.
        Thus we can just use str() to serialize them.
        To detect list we need to look at the split_to_list attribute of the ConfigVarDef and call str() on
        each element of the list, then join them with the separator used to create the list.
        If for_file is True we serialize only variables that have no_conffile_search set to False and
        used file_key as the key in the dict, otherwise the key is the variable name.
        """
        # we are going to take advantage of the AttrDict to accept dotted keys in order to create nested dicts.
        dumper = AttrDict()
        for var in vardef_to_serialize:
            if for_file and not var.flags.no_conffile_search:
                key = var.file_key
            elif not for_file:
                key = var.name
            else:
                continue

            sep = var.split_to_list if isinstance(var.split_to_list, str) else ","
            if var.split_to_list:
                val_list = [str(v) for v in self[var.name]]
                val = sep.join(val_list)
            else:
                val = str(self[var.name])

            dumper[key] = val

        return cast(dict[str, Any], dumper.to_dict())

    def store_conf_infile(
        self,
        file: Union[Path, str],
        *,
        namespaces: Sequence[str] = [],
        scopes: Sequence[str] = [],
        type: str = "yaml",
    ) -> None:
        """
        Store the current configuration in a file by merging the variables in the file with the one in this Appconfig object.
        The method writes the variables in a file.
        The variables to write must be in one of the specified namespaces and have one of the specified scopes.

        Args:
            file (Path | str ): The path to the file where to write the configuration.
            namespaces (Sequence[str], optional): A list of namespaces dotted names to filter the variables to write. Defaults to [].
            scopes (Sequence[str], optional): A list of scopes to filter the variables to write. Defaults to [].
            type (str, optional): The type of the file to write. Can be "yaml", "json" or "toml". Defaults to "yaml".
        """

        def get_vars_to_dump() -> Iterable[ConfigVarDef]:
            """
            loop over variable definitions to select the variables to dump. we use a separate function
            in ordre to return a generator to avoid double list traversal
            """
            for var in self._config_var_defs.values():

                # check namespaces filter if any
                if namespaces:
                    # if the variable is not in one of the requested namespaces skip it
                    if not any(var.name.startswith(prefix) for prefix in namespaces):
                        continue

                # check scope filter if any
                if scopes:
                    # if the variable is specific to some scopes check if one of them is in the requested scopes
                    if var.scopes is not None:
                        if not any(ctx in var.scopes for ctx in scopes):
                            continue

                # check if the variable should be in the file
                # redondant because file_key is None only if no_conffile_search is True, but useful for type checker that knwos that file_key is not None.
                if var.flags.no_conffile_search or var.file_key is None:
                    continue

                yield var

        if type not in ["json", "yaml", "toml"]:
            raise ValueError(
                f"Unknown type {type}. Supported types are 'json', 'toml' and 'yaml'."
            )
        data: dict[str, Any] = {}

        data = self._serialize_var_defs(get_vars_to_dump())
        if isinstance(file, str):
            file = Path(file)

        # if possible we need to read the file in order to inject our data into it and
        # not remove data in the file that we don't have in configuration.
        if file.exists() and os.access(file, os.R_OK):
            # file exists and is readable, we open it in read mode to load existing data
            with file.open("r", encoding="utf-8") as f:
                if type == "json":
                    existing_data = json.load(f)
                elif type == "yaml":
                    existing_data = yaml.safe_load(f)
                else:
                    existing_data = toml.load(f)
            # merge existing data with new data
            existing_data.update(data)
            data = existing_data
        if not file.parent.exists():
            file.parent.mkdir(parents=True, exist_ok=True)

        with file.open("w", encoding="utf-8") as f:
            if type == "json":
                json.dump(data, f, indent=4)
            elif type == "yaml":
                yaml.safe_dump(data, f)
            else:
                toml.dump(data, f)

    def serialize(self) -> dict[str, Any]:
        """Serialize the AppConfig to a dictionary."""
        return self._serialize_var_defs(self._config_var_defs.values())

    def __repr__(self) -> str:
        return str(self.to_dict())

    def __str__(self) -> str:
        return json.dumps(self.serialize(), indent=2, sort_keys=True, default=str)



from __future__ import annotations

import logging
from collections.abc import Iterator, Mapping
from typing import Any, Optional


class AttrDict:
    """
    class which properties can be accessed like a dict or like a propertie (.x) and vis/versa:
    if used like a Dict to set a value , if the key contains dots (.) it creates a hierarchy of AttrDict object.

    Takes advantage of the internal __dict__ attribute of a class instance where are stored all the instance attributes.
    The example bellow show that we can set and read attribute with the dot notation but also through the __dict__ attribute:

    class tt:
    def __init__(self):
        self.a =2

    vt = tt()
    print(vt.__dict__)      # {'a': 2}
    vt.b = 3
    print(vt.__dict__)      #  {'a': 2, 'b': 3}
    vt.__dict__["c"] = 4
    print(vt.__dict__)      # {'a': 2, 'b': 3, 'c': 4}


    class Test(AttrDict):
        def __init__(self):
            self.var1 = "var1"
            self.__dict__["vardict"] = "vardict"
    t = Test()
    print (f"t['vardict'] = {t['vardict']}") # t['vardict'] = vardict
    print (f"t.vardict = {t.vardict}")       # t.vardict = vardict
    print (f"t['var1'] = {t['var1']}")       # t['var1'] = var1
    print (f"t.var1 = {t.var1}")             # t.var1 = var1

    t.var2 = "var2"
    print (f"t['var2'] = {t['var2']}")       # t['var2'] = var2
    print (f"t.var2 = {t.var2}")             # t.var2 = var2

    t["var3"] = "var3"
    print (f"t['var3'] = {t['var3']}")       # t['var3'] = var3
    print (f"t.var3 = {t.var3}")             #  t.var3 = var3

    t["a1.a2"] = "complexval"
    print (f"t['a1'] = {t['a1']}")           # t['a1'] = {'a2': 'complexval'}
    print (f"t.a1 = {t.a1}")                 # t.a1 = {'a2': 'complexval'}
    """

    _logger = logging.getLogger(__name__)

    def __init__(self, init: Optional[Mapping[Any, Any]] = None) -> None:
        if init is not None:
            self.__dict__.update(init)

    def __contains__(self, key: Any) -> bool:
        try:
            eval(f"self['{key}']")
        except Exception:
            return False
        return True

    def __getitem__(self, key: Any) -> Any:
        """
        Called to implement evaluation of self[key]
        attributes with dots in their name are read in sub AttrDict Objects.
        """
        if isinstance(key, slice):
            raise KeyError("slices are not supported")
        parts = key.split(".")
        curpart = parts[0]

        if len(curpart.strip()) == 0:
            raise AttributeError(f"bad attribute name: {key}")
        elif len(parts) > 1:
            if curpart in self.__dict__:
                sub_attr = self.__dict__[curpart]
            else:
                raise AttributeError(f"key {curpart} not set")
            return sub_attr[".".join(parts[1:])]
        else:
            return self.__dict__[curpart]

    def __setitem__(self, key: Any, value: Any) -> None:
        """
        Called to implement assignment to self[key].
        attributes with dots in their name are put in sub AttrDict Objects.
        """
        if isinstance(key, slice):
            raise KeyError("slices are not supported")
        parts = key.split(".")
        curpart = parts[0]

        if len(curpart.strip()) == 0:
            raise AttributeError(f"bad attribute name: {key}")
        elif len(parts) > 1:
            if curpart in self.__dict__:
                sub_attr = self.__dict__[curpart]
            else:
                sub_attr = AttrDict()
                self.__dict__[curpart] = sub_attr
            sub_attr[".".join(parts[1:])] = value
        else:
            self.__dict__[curpart] = value

    def __delitem__(self, key: Any) -> None:
        """Called to implement deletion of self[key]."""
        del self.__dict__[key]

    def __len__(self) -> int:
        """Called to implement the built-in function len()"""
        return len(self.__dict__)

    def __iter__(self) -> Iterator:
        """This method is called when an iterator is required for a container."""
        return self.__dict__.__iter__()

    def update(self, other: Mapping[Any, Any], override: bool = False) -> None:
        if override:
            res = self.__dict__ | other  # type: ignore
        else:
            res = other | self.__dict__  # type: ignore
        self.__dict__ = res

    def to_dict(self) -> dict[Any, Any]:
        """
        Return a copy of the AttrDict as a dictionary.
        This is a recursive copy: nested AttrDict are also converted to dict.
        """
        d = {}
        for k, v in self.__dict__.items():
            if issubclass(type(v), AttrDict):
                d[k] = v.to_dict()
            else:
                d[k] = v
        return d

    def has_value(self, key: str | None) -> bool:
        """
        Check if the AttrDict has a value for the given key.
        The key can be a dotted path to access nested AttrDict.
        """
        if key is None:
            return False
        try:
            _ = self[key]
            return True
        except Exception:
            return False

    def _repr_with_ident(self, indent: int) -> list[str]:
        res = []
        indent_str = "\t" * indent
        for n, v in self.__dict__.items():
            if issubclass(self.__class__, type(v)):
                res.append(f"{indent_str}{n} :")
                res.extend(v._repr_with_ident(indent + 1))
            else:
                res.append("{}{} : {}".format(indent_str, n, v))
        return res

    def __repr__(self) -> str:
        res = self._repr_with_ident(0)
        return "\n".join(res)

    def clone(self) -> AttrDict:
        return AttrDict(self.__dict__)



from __future__ import annotations

import logging
import os
from typing import Any, Callable, Iterable, Mapping, Optional, Sequence, cast
from .config_var import ConfigVarDef, Flags
from .attr_dict import AttrDict
from .utils import AppConfigException, Undefined, _undef, is_sequence, PathType
from .var_substitution import render_template, resolve_candidate_list
from pydantic import TypeAdapter, ValidationError
from pathlib import Path


class Solver:
    """Class responsible for solving configuration variables values
    according to the defined search order and applying transformations and validations."""

    _logger = logging.getLogger(__name__)
    _local_flags: Flags
    _values_map: Mapping[str, Any]
    """ The values provided directly to the AppConfig instance."""
    _conffile_values_map: Optional[Mapping[str, Any]] = None
    """ The values found in the configuration file."""
    _eval_runner: Callable[[Callable[[str, Any, Any], Any], str, Any], Any]
    """ The function that will run the evaluation of callables for default values and transformations."""
    _var_solved: AttrDict
    """ currently solved config var. Used to handle dependencies between variables and variable substitution """

    def __init__(
        self,
        *,
        flags: Flags,
        values_map: Mapping[str, Any],
        eval_runner: Callable[[Callable[[str, Any, Any], Any], str, Any], Any],
        var_solved: AttrDict,
        conffile_values_map: Optional[Mapping[str, Any]] = None,
    ) -> None:
        self._local_flags = flags
        self._values_map = values_map
        self._conffile_values_map = conffile_values_map
        self._eval_runner = eval_runner
        self._var_solved = var_solved

    def set_conffile_values_map(
        self, conffile_values_map: Optional[Mapping[str, Any]]
    ) -> None:
        """
        Set or update the configuration file values map.
        Allow to use the same solver with ou without configuration file.
        """
        self._conffile_values_map = conffile_values_map

    @classmethod
    def _find_var_in_dict(
        cls, where: Mapping[Any, Any], var_name: str
    ) -> Any | Undefined:
        """
        Search for var_name in a dictionary.
        var_name contains dict keys separated by a dot (.) ie key1.key2.key3
        The left side key key1 is search in the where directory, if it is found
        and it is a dictionary then key2 is search in it and
        all key are searched in sequence.
        Returns _undef if not found, allowing to distinguish between None (explicit null) and missing.
        """
        cur_dic: Mapping[Any, Any] = where
        for key in var_name.split("."):
            # logger = logging.getLogger(f"{cls.__module__}.{cls.__name__}")
            # logger.debug(f"searching for key: {key} in dict : {cur_val}")
            if key in cur_dic:
                cur_dic = cur_dic[key]
                # logger.debug(f"Found key: {key} = {cur_val}")
            else:
                return _undef

        return cur_dic

    def resolve_confvar(
        self, *, var_to_solve: ConfigVarDef, scopes: Optional[str | Iterable[str]]
    ) -> Any:
        """
        Assign a value to the given ConfigVarDef according to the search order:
        1. values mapping
        2. environment variables
        3. configuration file
        4. default value
        Applies transformations and validations as specified in the ConfigVarDef.
        If its a Path variable, process relative paths, create directories if needed.
        """
        val_found: Path | str | None | list[Any] | Undefined = _undef
        found = False

        cur_flags = self._local_flags.merge(inplace=False, other=var_to_solve.flags)
        assert cur_flags is not None

        # if we are resolving the variables in some specific scopes (ie parameter scope has value)
        # we check if the variable belongs to one of these scopes and if not we skip it, in this
        # case the variable is said to be "out of scope" and has no existence.
        if scopes is None and var_to_solve.scopes is not None:
            return _undef
        # if we are resolving the variable in some specific scopes (ie parameter scope has value) and the variable
        # has also some scopes defined we check if there is at least one scope in common
        if scopes is not None and var_to_solve.scopes is not None:
            test_ctx = []
            # put True in test_ctx for each scope that matches, False otherwise
            if is_sequence(scopes):
                test_ctx = [c in var_to_solve.scopes for c in scopes]
            else:
                test_ctx = [scopes in var_to_solve.scopes]

            # test if there is any True in test_ctx which means that at least one scope matches
            if not any(test_ctx):
                return _undef
        # if the variable has no scope and we are resolving in some specific scopes this means that
        # the variable is common to all scopes so we process them but only once except if the flag
        # allow_override is set to True
        if var_to_solve.scopes is None and scopes is not None:
            if not cur_flags.allow_override and var_to_solve.name in self._var_solved:
                return _undef

        #  Searching variable value in the values mapping
        if (
            self._values_map is not None
            and not cur_flags.no_value_search
            and var_to_solve.value_key is not None
        ):
            ctx = self._var_solved.to_dict()
            # First Render value_key if needed

            def value_checks(k: str) -> Any:
                """check if the key k exists in the values_map and return its value or _undef"""
                if cur_flags.click_key_conversion:
                    k = k.lower().replace("-", "_")
                return _undef if k not in self._values_map else self._values_map[k]

            value_checked = resolve_candidate_list(
                candidates=var_to_solve.value_key,
                context=ctx,
                check_exists=lambda k: value_checks(k),
            )
            # value_checked can be None meaning that the key exists in values_map with a None value which is the value to use
            if value_checked != _undef:
                # We found a key that exists in values_map and the checker has returned its value that can be None.
                val_found = value_checked
                found = True
                self._logger.debug(
                    f"{var_to_solve.name} -> Found in Values = {val_found}"
                )

        # Searching variable value in the environment variables
        if (
            not found
            and not cur_flags.no_env_search
            and var_to_solve.env_name is not None
        ):
            ctx = self._var_solved.to_dict()

            def check_env(k: str) -> Any:
                """check if the key k exists in the environment variables and return its value or _undef"""
                res = os.getenv(k)
                return _undef if res is None else res

            value_checked = resolve_candidate_list(
                candidates=var_to_solve.env_name, context=ctx, check_exists=check_env
            )
            # in env None means not found (there is no _undef in this case and the value can't be None)
            if value_checked != _undef:
                val_found = value_checked
                found = True
                self._logger.debug(f"{var_to_solve.name} -> Found in Env = {val_found}")

        #  Searching variable value in the configuration file
        if (
            not found
            and not cur_flags.no_conffile_search
            and self._conffile_values_map is not None
            and var_to_solve.file_key is not None
        ):
            ctx = self._var_solved.to_dict()

            def check_file(k: str) -> Any:
                """check if the key k exists in configuration file and return its value or _undef"""
                if self._conffile_values_map is None:
                    return False
                return self._find_var_in_dict(self._conffile_values_map, k)

            value_checked = resolve_candidate_list(
                candidates=var_to_solve.file_key, context=ctx, check_exists=check_file
            )
            # None is a valid value from config file, _undef means not found
            if value_checked != _undef:
                val_found = value_checked
                found = True
                self._logger.debug(
                    f"{var_to_solve.name} -> Found in Configuration File = {val_found}"
                )

        #  Setting variable value to the default value if defined.
        if (
            var_to_solve.name not in self._var_solved
            and not found
            and var_to_solve.default is not _undef
        ):
            if var_to_solve.default is not None and callable(var_to_solve.default):
                val_found = self._eval_runner(
                    var_to_solve.default, var_to_solve.name, None
                )
            else:
                val_found = render_template(
                    var_to_solve.default, self._var_solved.to_dict()
                )
            found = True
            self._logger.debug(
                f"{var_to_solve.name} -> Found in Default Value = {val_found}"
            )

        # Raise exception if no value was found and variable is mandatory and if we have not
        # already a value for this variable (in case of override with scopes for example)
        if (
            not found
            and var_to_solve.mandatory
            and var_to_solve.name not in self._var_solved
        ):
            raise AppConfigException(
                f"No value for var {var_to_solve.name} in scope {scopes}"
            )

        if found:
            if isinstance(val_found, Undefined):
                raise AssertionError(
                    "Internal error: val_found is _undef while found=True"
                )
            # spliting lists
            if (
                var_to_solve.split_to_list
                and val_found is not None
                and isinstance(val_found, str)
            ):
                sep: str = (
                    ","
                    if isinstance(var_to_solve.split_to_list, bool)
                    else var_to_solve.split_to_list
                )
                val_found = val_found.split(sep)

            # Transform the variable value if specified.
            if var_to_solve.transform is not None:
                if isinstance(var_to_solve.transform, str):
                    ctx = self._var_solved.to_dict()
                    ctx["value"] = val_found
                    val_transfo = render_template(var_to_solve.transform, ctx)
                else:
                    val_transfo = self._eval_runner(
                        var_to_solve.transform,
                        var_to_solve.name,
                        val_found,
                    )
                self._logger.debug(
                    f"{var_to_solve.name} -> Value Transformed: {val_found} => {val_transfo}"
                )
                val_found = val_transfo

            # Validate and cast the variable value using pydantic TypeAdapter
            if var_to_solve._type_info is not None and val_found is not None:
                try:
                    ta = TypeAdapter(var_to_solve._type_info)
                    val_found = ta.validate_python(val_found)
                except ValidationError as e:
                    raise AppConfigException(
                        f"Validation failed for var {var_to_solve.name}: {e}"
                    )

            # Make special treatment for paths
            if (
                not var_to_solve.no_dir_processing
                and var_to_solve._type_info == Path
                and val_found is not None
            ):
                new_val: list[Path] | Path | None = None
                if is_sequence(val_found):
                    # we need to consider the case where the variable is a list of Path
                    new_val = []
                    for cval in cast(Iterable, val_found):
                        res = self._process_paths(
                            value=cval, var=var_to_solve, var_solved=self._var_solved
                        )
                        new_val.append(res)
                else:
                    res = self._process_paths(
                        value=val_found, var=var_to_solve, var_solved=self._var_solved
                    )
                    new_val = res
                val_found = new_val

        elif var_to_solve.name not in self._var_solved:
            # in case of overiding and when on subsenquent call no value were found we don't want to erase a value
            # found on a previous run. This can happen when some global flags are overriden or for variable without
            # context.
            val_found = None

        return val_found

    def _process_paths(
        self, *, value: Any, var: ConfigVarDef, var_solved: AttrDict
    ) -> Path:
        """
        Run the path processing for a single Path variable.
        """
        var_value: Path
        if isinstance(value, str):
            var_value = Path(value)
        elif isinstance(value, Path):
            var_value = value
        else:
            raise Exception("not a path")

        res: Path = var_value
        # First process the can_be_relative_to case.
        if var.can_be_relative_to is not None:

            if var_value.is_absolute():
                res = var_value
            else:
                # resolving the root directory from which to be relative to.
                can_be_relative = var.can_be_relative_to
                root_dir: Path
                if (
                    isinstance(can_be_relative, str)
                    and can_be_relative in var_solved
                    and var_solved[can_be_relative] is not None
                    and var_solved[can_be_relative] != _undef
                ):
                    root_dir = Path(var_solved[can_be_relative])
                else:
                    root_dir = Path(can_be_relative)
                res = root_dir.joinpath(var_value)

        # Expanding user home and resolving dots in path
        res = res.expanduser().resolve()

        # create directory if make_dirs is not None
        if var.make_dirs:
            if var.make_dirs == PathType.Dir:
                res.mkdir(parents=True, exist_ok=True)
            else:
                res.parent.mkdir(parents=True, exist_ok=True)
        return res



from .app_config import AppConfig
from .config_var import ConfigVarDef, configvar
from .click_option import click_option
from .utils import PathType, AppConfigException, NameSpace

__all__ = [
    "AppConfig",
    "AppConfigException",
    "ConfigVarDef",
    "NameSpace",
    "configvar",
    "click_option",
    "PathType",
]



from __future__ import annotations

from typing import Any, Callable

import click

from .app_config import ConfigVarDef
from .utils import Undefined, _undef


class StringOrUndefined(click.ParamType):
    name = "string"

    def convert(self, value: Any, param: Any, ctx: Any) -> Any:
        if value is _undef:
            return value
        return click.STRING.convert(value, param, ctx)


def click_option(config_var: Any, **kwargs: Any) -> Callable[[Any], Any]:
    """Wrapper for click.option with the metadata from ConfigVarDef."""
    if not isinstance(config_var, ConfigVarDef):
        raise TypeError(f"click_option expects a ConfigVarDef, got {type(config_var)}")

    # 1. Flag name (ex: my_var -> --my-var)
    param_decls = []

    keys = []
    if config_var.value_key is None:
        keys = [config_var.name]
    elif isinstance(config_var.value_key, str):
        keys = [config_var.value_key]
    else:
        keys = config_var.value_key

    for key in keys:
        if len(key) == 1:
            param_decls.append(f"-{key}")
        else:
            param_decls.append(f"--{key.replace('_', '-').lower()}")

    # Used as destination in the dict returned by click
    param_name = keys[0]

    # 2. Documentation
    if "help" not in kwargs and config_var.help:
        kwargs["help"] = config_var.help

    # 3. Strict constraint: Always strings
    # We overwrite any passed type to ensure that resolve_vars receives a string.
    # We use a custom type to handle the `_undef` default value without converting it to string.
    kwargs["type"] = StringOrUndefined()

    # We set default to `_undef` to distinguish between "option not provided" (_undef)
    # and "option provided with no value" (which shouldn't happen with string) or None.
    # AppConfig handles `_undef` as "missing value" and will look for other sources (env, file, default).
    kwargs["default"] = _undef

    # 4. Display default without applying it
    if (
        "show_default" not in kwargs
        and config_var.default is not Undefined
        and not callable(config_var.default)
    ):
        # We must convert to string otherwise click interprets bool/int as flags (True/False)
        # which tell it "show the default" (which is None here), so it displays nothing.
        kwargs["show_default"] = str(config_var.default)

    # Note: We do NOT pass 'envvar' nor 'default'

    # We force the destination name to match the key expected by the1conf
    return click.option(*param_decls, param_name, **kwargs)



from __future__ import annotations

import copy
from collections import OrderedDict
from inspect import getsource
from typing import Any, Iterable, cast
from .config_var import ConfigVarDef
from .utils import get_attribute_docstrings, NameSpace


class ConfigVarCollector:
    """
    Collector class to gather ConfigVarDef from AppConfig classes and their nested NameSpace classes.

    This class helps in aggregating configuration variables from various sources during the class creation process.
    It supports two main collection strategies:
    1. Parsing the class definition directly (useful for mixins or the root class).
    2. Inheriting already collected variables from a parent `AppConfig` class (optimization).

    This class can be used as a context manager to ensure proper resource management.
    """

    _collected_vars: OrderedDict[str, ConfigVarDef]
    _auto_prolog_var_names: list[str]

    def __enter__(self):
        """Return self to be used in the with statement"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Cleanup resources on exit"""
        self.close()

    def __init__(self) -> None:
        """Create a new ConfigVarCollector instance."""
        self._collected_vars = OrderedDict()
        self._auto_prolog_var_names = []

    def close(self) -> None:
        """
        Clean up resources if needed.
        """
        self._collected_vars.clear()
        self._auto_prolog_var_names.clear()

    def _collect_config_vars(
        self, owner: type, prefix: str = ""
    ) -> Iterable[ConfigVarDef]:
        """
        Traverses ConfigVarDef defined in the owner class and its nested NameSpace classes.
        Search for ConfigVarDef in the owner class __dict__ and yield them.
        Attach the docstrings as help if help is not already defined and a docstring exists
        for the attribute.
        change its name by adding the namespace it is defined in as a prefix.
        """
        # extract docstrings from the class source code
        docstrings = get_attribute_docstrings(owner)

        # Inspect the class __dict__ to find ConfigVarDef defined in this class
        for attr_name, value in owner.__dict__.items():
            if isinstance(value, ConfigVarDef):
                # Clone to update name with prefix
                new_var = value

                # If help is missing, try to use the docstring
                if not new_var.help and attr_name in docstrings:
                    new_var._help = docstrings[attr_name]

                if prefix:
                    # We need to access _name directly because we don't want to trigger the
                    # descriptor __get__ method
                    original_name = getattr(value, "_name", attr_name)
                    new_var._name = f"{prefix}{original_name}"

                yield new_var

            elif (
                isinstance(value, type)
                and issubclass(value, NameSpace)
                and value is not NameSpace
            ):
                # Recurse into nested NameSpace classes
                yield from self._collect_config_vars(
                    value, prefix=f"{prefix}{attr_name}."
                )

    def collect_on_class(self, owner: type) -> None:
        """
        Collect configuration variables by inspecting the class `__dict__` and its nested `NameSpace`s.

        This method is used when the class does not have a `_config_var_defs` attribute yet,
        typically for mixins (like AutoProlog) or during the initialization of the root `AppConfig` class.

        It performs a deep scan:
        - Iterates over the class attributes to find `ConfigVarDef` instances.
        - Extracts docstrings to use as help text if needed.
        - Recurses into nested `NameSpace` classes to collect variables defined within them.

        Any variable found overrides an existing variable with the same name in the collection.
        """
        for vardef in self._collect_config_vars(owner):
            attr_name = getattr(vardef, "_name", None)
            if attr_name:
                if attr_name in self._collected_vars:
                    # Override previous definition
                    self._collected_vars.pop(attr_name)
                self._collected_vars[attr_name] = vardef
                if attr_name and vardef.auto_prolog:
                    if attr_name in self._auto_prolog_var_names:
                        # in this case we need to move the var_name to the end of the list to preserve the order
                        self._auto_prolog_var_names.remove(attr_name)
                    self._auto_prolog_var_names.append(attr_name)

    def collect_in_config_var_defs(self, owner: type) -> None:
        """
        Collect configuration variables from the `_config_var_defs` attribute of the owner class.

        This method is used during inheritance when the parent class (`owner`) is already a subclass
        of `AppConfig`. Since the parent has already computed its configuration variables (stored in
        `_config_var_defs`), we can simply copy them instead of re-scanning the class.

        This is much faster than `collect_on_class` and preserves the resolution order established
        in the parent class.

        Any variable found overrides an existing variable with the same name in the collection.
        """
        base_config_var_defs = getattr(owner, "_config_var_defs", None)
        if not base_config_var_defs:
            return

        assert isinstance(base_config_var_defs, OrderedDict)

        for var_name, var_def in cast(
            OrderedDict[str, ConfigVarDef], base_config_var_defs
        ).items():
            # base_config_var_defs is the _config_var_defs dict. Loops on each of its elements and store them in the collected list
            # For each one read its name. If the variable name is already in the collected list, it means that it has been
            # defined in a parent class of the current inspected base class. So we remove the old definition and add the new one.
            if var_name:
                if var_name in self._collected_vars:
                    # Override previous definition
                    self._collected_vars.pop(var_name)
                self._collected_vars[var_name] = var_def
                if var_name and var_def.auto_prolog:
                    if var_name in self._auto_prolog_var_names:
                        # in this case we need to move the var_name to the end of the list to preserve the order
                        self._auto_prolog_var_names.remove(var_name)
                    self._auto_prolog_var_names.append(var_name)

    def get_collected_vars(self) -> OrderedDict[str, ConfigVarDef]:
        """
        Get the collected configuration variables.
        """
        return self._collected_vars.copy()

    def get_auto_prolog_var_names(self) -> list[str]:
        """
        Get only the collected configuration variables that have auto_prolog enabled.
        """
        return self._auto_prolog_var_names.copy()


class AppConfigMeta(type):
    """
    Metaclass for AppConfig to handle initialization of the root class itself,
    specifically to collect configuration variables from mixins that are not AppConfig subclasses.
    """

    def __init__(cls, name: str, bases: tuple[type, ...], attrs: dict[str, Any]) -> None:
        """
        Initialize the class object.

        This method is part of the Python metaclass protocol. It is invoked after the class object `cls`
        has been created (usually via `type.__new__`). It effectively acts as a constructor for the class
        object itself.

        Args:
            name (str): The name of the class being created.
            bases (tuple[type, ...]): A tuple containing the base classes of the class being created.
            attrs (dict[str, Any]): A dictionary containing the class namespace (attributes and methods)
                                  populated during the execution of the class body.
                                  Keys are the names of the attributes (strings) and values are the
                                  attribute values (methods, class variables, properties, etc.).
        """
        super().__init__(name, bases, attrs)

        # Only perform root initialization for the AppConfig class itself.
        # Subclasses use the standard __init_subclass__ mechanism.
        if name == "AppConfig":
            with ConfigVarCollector() as collector:
                for base in reversed(cls.__mro__[1:-1]):
                    # For the root class initialization we want to scan all parents.
                    # We want to include variables from mixins like AutoProlog.
                    # Since AppConfig is the root for this mechanism, its parents don't have _config_var_defs.
                    # In the mro list AppConfig is always at index 0, so we can skip it safely and the last item is always object which we can also skip.

                    collector.collect_on_class(base)
                # Add own vars
                collector.collect_on_class(cls)

                cls._config_var_defs = collector.get_collected_vars()
                cls._autoprolog_var_names = collector.get_auto_prolog_var_names()



from __future__ import annotations

from typing import Optional


class Flags:
    """
    Flags that control how to search for variable values in AppConfig.
    These flags can be set globally in the AppConfig object or locally when calling resolve_vars() or
    per config variables.

    Not None flag values override other values with the following precedence:
        ConfigVariableDef > resolve_vars() parameters > AppConfig global flags.
    Which means that a not None value in a ConfVarDef directive will overrride a global flag value defined in AppConfig. And also that a None value in a ConfVarDef directive
    will not override the value set in resolve_vars() parameters or an AppConfig global flags.

    Attributes:
        no_env_search (bool|None): don't search values in the environment.
        no_value_search (bool|None): don't search values in the values dict.
        no_conffile_search (bool|None): don't search values in the configuration file.
        no_search (bool|None): don't search values in any location.
        click_key_conversion (bool|None): use the variable name converted to lower case and with dashes converted to underscores as
            the key to search in values.
        allow_override (bool|None): allow overriding variable in the configuration when calling resolve_vars.
        mandatory (bool|None): raise exception when no value is found for a variable. Default is True.
    """

    _no_env_search: bool | None
    _no_value_search: bool | None
    _no_conffile_search: bool | None
    _no_search: bool | None
    _click_key_conversion: bool | None
    _allow_override: bool | None

    def __init__(
        self,
        *,
        no_env_search: Optional[bool] = None,
        no_value_search: Optional[bool] = None,
        no_conffile_search: Optional[bool] = None,
        no_search: Optional[bool] = None,
        click_key_conversion: Optional[bool] = None,
        allow_override: Optional[bool] = None,
    ) -> None:
        self._no_env_search = no_env_search
        self._no_value_search = no_value_search
        self._no_conffile_search = no_conffile_search
        self._no_search = no_search
        if self._no_search:
            self._no_env_search = True
            self._no_value_search = True
            self._no_conffile_search = True
        self._click_key_conversion = click_key_conversion
        self._allow_override = allow_override

    def merge(self, *, inplace: bool = False, other: Flags) -> Flags | None:
        """
        Merge the current flags with the given flags.
        If inplace is True then modify the current object, otherwise return a new Flags object.

        Flags given as parameters have precedence over the current object values if they are not None.
        """
        cur_no_search: Optional[bool]
        cur_no_env_search: Optional[bool]
        cur_no_value_search: Optional[bool]
        cur_no_conffile_search: Optional[bool]
        cur_click_key_conversion: Optional[bool]
        cur_allow_override: Optional[bool]

        cur_no_search = (
            other._no_search if other._no_search is not None else self._no_search
        )
        if cur_no_search:
            cur_no_env_search = True
            cur_no_value_search = True
            cur_no_conffile_search = True
        else:
            cur_no_env_search = (
                other._no_env_search
                if other._no_env_search is not None
                else self._no_env_search
            )
            cur_no_value_search = (
                other._no_value_search
                if other._no_value_search is not None
                else self._no_value_search
            )
            cur_no_conffile_search = (
                other._no_conffile_search
                if other._no_conffile_search is not None
                else self._no_conffile_search
            )
        cur_click_key_conversion = (
            other._click_key_conversion
            if other._click_key_conversion is not None
            else self._click_key_conversion
        )
        cur_allow_override = (
            other._allow_override
            if other._allow_override is not None
            else self._allow_override
        )
        if not inplace:
            return Flags(
                no_env_search=cur_no_env_search,
                no_value_search=cur_no_value_search,
                no_conffile_search=cur_no_conffile_search,
                no_search=cur_no_search,
                click_key_conversion=cur_click_key_conversion,
                allow_override=cur_allow_override,
            )
        else:
            self._no_env_search = cur_no_env_search
            self._no_value_search = cur_no_value_search
            self._no_conffile_search = cur_no_conffile_search
            self._no_search = cur_no_search
            self._click_key_conversion = cur_click_key_conversion
            self._allow_override = cur_allow_override
            return None

    @property
    def no_env_search(self) -> bool:
        return False if self._no_env_search is None else self._no_env_search

    @property
    def no_value_search(self) -> bool:
        return False if self._no_value_search is None else self._no_value_search

    @property
    def no_conffile_search(self) -> bool:
        return False if self._no_conffile_search is None else self._no_conffile_search

    @property
    def no_search(self) -> bool:
        return False if self._no_search is None else self._no_search

    @property
    def click_key_conversion(self) -> bool:
        return False if self._click_key_conversion is None else self._click_key_conversion

    @property
    def allow_override(self) -> bool:
        return False if self._allow_override is None else self._allow_override



from __future__ import annotations

import sys
from pathlib import Path
from typing import Literal, Optional, Any

from .config_var import configvar
from .utils import PathType


def _get_os_type() -> str:
    if sys.platform.startswith("win"):
        return "windows"
    elif sys.platform.startswith("linux"):
        return "linux"
    elif sys.platform.startswith("darwin"):
        return "macos"
    return "unknown"


class AutoProlog:
    """
    Automatic configuration for standard paths and environment variables.
    This class defines configuration variables that will be accessible in any
    subclass of AppConfig, providing sensible defaults based directories, Os type
    and execution stage.

    Variables defined:
    - os_type: The operating system type (windows, linux, macos)
    - user_home: The user's home directory, as returned by Path.home()
      Must be used instead of '~' to ensure cross platform compatibility.
    - xdg_data_home: Base directory for user specific data files (XDG standard)
    - xdg_config_home: Base directory for user specific configuration files (XDG standard)
    - xdg_cache_home: Base directory for user specific non-essential data files (XDG standard)
    - xdg_state_home: Base directory for user specific state files (XDG standard)
    - app_data: Application Data directory on Windows
    - app_config: Application Config directory on Windows
    - app_cache: Application Cache directory on Windows
    - data_home: OS independent application data directory
    - config_home: OS independent application configuration directory
    - cache_home: OS independent application cache directory
    - exec_stage: The execution stage, default to the string "dev", no restriction on its value
        to let the user define its own stages.

    Any of this variable can be overridden by a redefinition on a subclasse of AppConfig.
    """

    os_type: Literal["windows", "linux", "macos", "unknown"] = configvar(
        default=lambda _, __, ___: _get_os_type(),
        no_search=True,
        help="The operating system type (windows, linux, macos)",
        auto_prolog=True,
    )

    exec_stage: Any = configvar(
        default=None,
        type_info=str,
        env_name=["EXEC_STAGE", "STAGE", "ENV"],
        value_key=["exec_stage", "stage", "env"],
        help="The execution stage (dev, prod, test)",
        click_key_conversion=True,
        allow_override=True,
        auto_prolog=True,
    )

    user_home: Path = configvar(
        default=lambda _, __, ___: Path.home(),
        no_search=True,
        help="The user's home directory",
        auto_prolog=True,
    )

    # XDG Standards
    xdg_data_home: Optional[Path] = configvar(
        default=lambda _, c, __: (
            c.user_home / ".local" / "share" if c.os_type != "windows" else None
        ),
        env_name="XDG_DATA_HOME",
        help="Base directory for user specific data files",
        no_value_search=True,
        no_conffile_search=True,
        auto_prolog=True,
    )

    xdg_config_home: Optional[Path] = configvar(
        default=lambda _, c, __: (
            c.user_home / ".config" if c.os_type != "windows" else None
        ),
        env_name="XDG_CONFIG_HOME",
        help="Base directory for user specific configuration files",
        no_value_search=True,
        no_conffile_search=True,
        auto_prolog=True,
    )

    xdg_cache_home: Optional[Path] = configvar(
        default=lambda _, c, __: (
            c.user_home / ".cache" if c.os_type != "windows" else None
        ),
        env_name="XDG_CACHE_HOME",
        help="Base directory for user specific non-essential data files",
        no_value_search=True,
        no_conffile_search=True,
        auto_prolog=True,
    )

    xdg_state_home: Optional[Path] = configvar(
        default=lambda _, c, __: (
            c.user_home / ".local" / "state" if c.os_type != "windows" else None
        ),
        env_name="XDG_STATE_HOME",
        help="Base directory for user specific state files",
        no_value_search=True,
        no_conffile_search=True,
        auto_prolog=True,
    )

    # Windows specific variables
    app_data: Optional[Path] = configvar(
        default=lambda _, c, __: (
            c.user_home / "AppData" if c.os_type == "windows" else None
        ),
        env_name="APP_DATA",
        help="Application Data directory on Windows",
        no_value_search=True,
        no_conffile_search=True,
        auto_prolog=True,
    )

    app_config: Optional[Path] = configvar(
        default=lambda _, c, __: c.app_data / "Config" if c.app_data else None,
        env_name="APP_CONFIG",
        help="Application Config directory on Windows",
        no_value_search=True,
        no_conffile_search=True,
        auto_prolog=True,
    )

    app_cache: Optional[Path] = configvar(
        default=lambda _, c, __: c.app_data / "Cache" if c.app_data else None,
        env_name="APP_CACHE",
        help="Application Cache directory on Windows",
        no_value_search=True,
        no_conffile_search=True,
        auto_prolog=True,
    )

    # Unified variables
    data_home: Path = configvar(
        default=lambda _, c, __: (
            c.app_data if c.os_type == "windows" else c.xdg_data_home
        ),
        click_key_conversion=True,
        help="OS independent application data directory",
        auto_prolog=True,
        allow_override=True,
    )

    config_home: Path = configvar(
        default=lambda _, c, __: (
            c.app_config if c.os_type == "windows" else c.xdg_config_home
        ),
        click_key_conversion=True,
        help="OS independent application configuration directory",
        auto_prolog=True,
        allow_override=True,
    )

    cache_home: Path = configvar(
        default=lambda _, c, __: (
            c.app_cache if c.os_type == "windows" else c.xdg_cache_home
        ),
        click_key_conversion=True,
        help="OS independent application cache directory",
        auto_prolog=True,
        allow_override=True,
    )
